// Class: ReadNeuralNetwork
// Automatically generated by MethodBase::MakeClass
//

/* configuration options =====================================================

#GEN -*-*-*-*-*-*-*-*-*-*-*- general info -*-*-*-*-*-*-*-*-*-*-*-

Method         : MLP::NeuralNetwork
TMVA Release   : 3.8.13        [198669]
ROOT Release   : 5.18/00       [332288]
Creator        : lista
Date           : Mon Feb  9 12:48:27 2009
Host           : Linux lxbuild063.cern.ch 2.6.9-78.0.1.EL.cernsmp #1 SMP Tue Aug 5 11:01:13 CEST 2008 i686 i686 i386 GNU/Linux
Dir            : /data1/home/lista/stat/03-TMVA
Training events: 200000


#OPT -*-*-*-*-*-*-*-*-*-*-*-*- options -*-*-*-*-*-*-*-*-*-*-*-*-

# Set by User:
Normalise: "True" [Normalise input variables]
NCycles: "200" [Number of training cycles]
HiddenLayers: "N+2,N" [Specification of hidden layer architecture]
NeuronInputType: "sum" [Neuron input function type]
LearningRate: "0.02" [ANN learning rate parameter]
DecayRate: "0.01" [Decay rate for learning parameter]
# Default:
D: "False" [use-decorrelated-variables flag (depreciated)]
VarTransform: "None" [Variable transformation method]
VarTransformType: "Signal" [Use signal or background events for var transform]
NbinsMVAPdf: "60" [Number of bins used to create MVA PDF]
NsmoothMVAPdf: "2" [Number of smoothing iterations for MVA PDF]
V: "False" [Verbose mode]
VerboseLevel: "Info" [Verbosity level]
H: "False" [Print classifier-specific help message]
CreateMVAPdfs: "False" [Create PDFs for classifier outputs]
TxtWeightFilesOnly: "True" [if True, write all weights as text files]
NeuronType: "tanh" [Neuron activation function type]
TrainingMethod: "BP" [Train with Back-Propagation (BP - default) or Genetic Algorithm (GA - slower and worse)]
TestRate: "10" [Test for overtraining performed at each #th epochs]
BPMode: "sequential" [Back-propagation learning mode: sequential or batch]
BatchSize: "-1" [Batch size: number of events/batch, only set if in Batch Mode, -1 for BatchSize=number_of_events]
##


#VAR -*-*-*-*-*-*-*-*-*-*-*-* variables *-*-*-*-*-*-*-*-*-*-*-*-

NVar 2
x                             x                                 'F'    [-25.9051780701,28.9163246155]
y                             y                                 'F'    [-10.0186662674,11.3229227066]


============================================================================ */

#include <vector>
#include <cmath>
#include <string>
#include <iostream>

#ifndef IClassifierReader__def
#define IClassifierReader__def

class IClassifierReader {

 public:

   // constructor
   IClassifierReader() {}
   virtual ~IClassifierReader() {}

   // return classifier response
   virtual double GetMvaValue( const std::vector<double>& inputValues ) const = 0;
};

#endif

class ReadNeuralNetwork : public IClassifierReader {

 public:

   // constructor
   ReadNeuralNetwork( std::vector<std::string>& theInputVars ) 
      : IClassifierReader(),
        fClassName( "ReadNeuralNetwork" ),
        fStatusIsClean( true ),
        fNvars( 2 ),
        fIsNormalised( true )
   {      
      // the training input variables
      const char* inputVars[] = { "x", "y" };

      // sanity checks
      if (theInputVars.size() <= 0) {
         std::cout << "Problem in class \"" << fClassName << "\": empty input vector" << std::endl;
         fStatusIsClean = false;
      }

      if (theInputVars.size() != fNvars) {
         std::cout << "Problem in class \"" << fClassName << "\": mismatch in number of input values: "
                   << theInputVars.size() << " != " << fNvars << std::endl;
         fStatusIsClean = false;
      }

      // validate input variables
      for (size_t ivar = 0; ivar < theInputVars.size(); ivar++) {
         if (theInputVars[ivar] != inputVars[ivar]) {
            std::cout << "Problem in class \"" << fClassName << "\": mismatch in input variable names" << std::endl
                      << " for variable [" << ivar << "]: " << theInputVars[ivar].c_str() << " != " << inputVars[ivar] << std::endl;
            fStatusIsClean = false;
         }
      }

      // initialize min and max vectors (for normalisation)
      fVmin[0] = -25.9051780700684;
      fVmax[0] = 28.9163246154785;
      fVmin[1] = -10.018666267395;
      fVmax[1] = 11.322922706604;

      // initialize input variable types
      fType[0] = 'F';
      fType[1] = 'F';

      // initialize constants
      Initialize();

   }

   // destructor
   virtual ~ReadNeuralNetwork() {
      Clear(); // method-specific
   }

   // the classifier response
   // "inputValues" is a vector of input values in the same order as the 
   // variables given to the constructor
   double GetMvaValue( const std::vector<double>& inputValues ) const;

 private:

   // method-specific destructor
   void Clear();

   // common member variables
   const char* fClassName;
   bool fStatusIsClean;

   const size_t fNvars;
   size_t GetNvar()           const { return fNvars; }
   char   GetType( int ivar ) const { return fType[ivar]; }

   // normalisation of input variables
   const bool fIsNormalised;
   bool IsNormalised() const { return fIsNormalised; }
   double fVmin[2];
   double fVmax[2];
   double NormVariable( double x, double xmin, double xmax ) const {
      // normalise to output range: [-1, 1]
      return 2*(x - xmin)/(xmax - xmin) - 1.0;
   }

   // type of input variable: 'F' or 'I'
   char   fType[2];

   // initialize internal variables
   void Initialize();
   double GetMvaValue__( const std::vector<double>& inputValues ) const;

   // private members (method specific)

   double ActivationFnc(double x) const;

   int fLayers;
   int fLayerSize[4];
   double fWeightMatrix0to1[5][3];   // weight matrix from layer 0 to 1
   double fWeightMatrix1to2[3][5];   // weight matrix from layer 1 to 2
   double fWeightMatrix2to3[1][3];   // weight matrix from layer 2 to 3

   double * fWeights[4];
};

inline void ReadNeuralNetwork::Initialize()
{
   // build network structure
   fLayers = 4;
   fLayerSize[0] = 3; fWeights[0] = new double[3]; 
   fLayerSize[1] = 5; fWeights[1] = new double[5]; 
   fLayerSize[2] = 3; fWeights[2] = new double[3]; 
   fLayerSize[3] = 1; fWeights[3] = new double[1]; 
   // weight matrix from layer 0 to 1
   fWeightMatrix0to1[0][0] = -0.213524276774962;
   fWeightMatrix0to1[1][0] = 3.66016148829324;
   fWeightMatrix0to1[2][0] = 6.94936422634046;
   fWeightMatrix0to1[3][0] = -0.873220328298971;
   fWeightMatrix0to1[0][1] = -5.64740983296147;
   fWeightMatrix0to1[1][1] = -2.52130036188141;
   fWeightMatrix0to1[2][1] = -2.09138202692104;
   fWeightMatrix0to1[3][1] = 6.23634121296057;
   fWeightMatrix0to1[0][2] = -2.86931384972127;
   fWeightMatrix0to1[1][2] = 0.176818000338653;
   fWeightMatrix0to1[2][2] = -4.46195542969536;
   fWeightMatrix0to1[3][2] = -2.19207532370122;
   // weight matrix from layer 1 to 2
   fWeightMatrix1to2[0][0] = -3.24202661856694;
   fWeightMatrix1to2[1][0] = -6.4233628066801;
   fWeightMatrix1to2[0][1] = 3.55895409490169;
   fWeightMatrix1to2[1][1] = 0.211259999298389;
   fWeightMatrix1to2[0][2] = -2.3931830006396;
   fWeightMatrix1to2[1][2] = 2.57236328003327;
   fWeightMatrix1to2[0][3] = 3.12561144474314;
   fWeightMatrix1to2[1][3] = 0.116945694086782;
   fWeightMatrix1to2[0][4] = -2.96275660966917;
   fWeightMatrix1to2[1][4] = -3.85859065674936;
   // weight matrix from layer 2 to 3
   fWeightMatrix2to3[0][0] = 1.02187175198735;
   fWeightMatrix2to3[0][1] = -0.106384209854834;
   fWeightMatrix2to3[0][2] = -0.0220788944598061;
}

inline double ReadNeuralNetwork::GetMvaValue__( const std::vector<double>& inputValues ) const
{
   if (inputValues.size() != (unsigned int)fLayerSize[0]-1) {
      std::cout << "Input vector needs to be of size " << fLayerSize[0]-1 << std::endl;
      return 0;
   }

   for (int l=0; l<fLayers; l++)
      for (int i=0; i<fLayerSize[l]; i++) fWeights[l][i]=0;

   for (int l=0; l<fLayers-1; l++)
      fWeights[l][fLayerSize[l]-1]=1;

   for (int i=0; i<fLayerSize[0]-1; i++)
      fWeights[0][i]=inputValues[i];

   // layer 0 to 1
   for (int o=0; o<fLayerSize[1]-1; o++) {
      for (int i=0; i<fLayerSize[0]; i++) {
         double inputVal = fWeightMatrix0to1[o][i] * fWeights[0][i];
         fWeights[1][o] += inputVal;
      }
      fWeights[1][o] = ActivationFnc(fWeights[1][o]);
   }
   // layer 1 to 2
   for (int o=0; o<fLayerSize[2]-1; o++) {
      for (int i=0; i<fLayerSize[1]; i++) {
         double inputVal = fWeightMatrix1to2[o][i] * fWeights[1][i];
         fWeights[2][o] += inputVal;
      }
      fWeights[2][o] = ActivationFnc(fWeights[2][o]);
   }
   // layer 2 to 3
   for (int o=0; o<fLayerSize[3]; o++) {
      for (int i=0; i<fLayerSize[2]; i++) {
         double inputVal = fWeightMatrix2to3[o][i] * fWeights[2][i];
         fWeights[3][o] += inputVal;
      }
   }

   return fWeights[3][0];
}

double ReadNeuralNetwork::ActivationFnc(double x) const {
   // hyperbolic tan
   return tanh(-x);
}
   
// Clean up
inline void ReadNeuralNetwork::Clear() 
{
   // nothing to clear
}
   inline double ReadNeuralNetwork::GetMvaValue( const std::vector<double>& inputValues ) const
   {
      // classifier response value
      double retval = 0;

      // classifier response, sanity check first
      if (!fStatusIsClean) {
         std::cout << "Problem in class \"" << fClassName << "\": cannot return classifier response"
                   << " because status is dirty" << std::endl;
         retval = 0;
      }
      else {
         if (IsNormalised()) {
            // normalise variables
            std::vector<double> iV;
            int ivar = 0;
            for (std::vector<double>::const_iterator varIt = inputValues.begin();
                 varIt != inputValues.end(); varIt++, ivar++) {
               iV.push_back(NormVariable( *varIt, fVmin[ivar], fVmax[ivar] ));
            }
            retval = GetMvaValue__( iV );
         }
         else {
            retval = GetMvaValue__( inputValues );
         }
      }

      return retval;
   }
